@@ -1,6 +1,6 @@
 /*
  * GeeXboX Valhalla: tiny media scanner API.
- * Copyright (C) 2009 Mathieu Schroeter <mathieu.schroeter@gamesover.ch>
+ * Copyright (C) 2009-2010 Mathieu Schroeter <mathieu.schroeter@gamesover.ch>
  *
  * This file is part of libvalhalla.
  *
@@ -29,6 +29,7 @@
 #include "valhalla_internals.h"
 #include "fifo_queue.h"
 #include "logs.h"
+#include "stats.h"
 #include "thread_utils.h"
 #include "event_handler.h"
 #include "database.h"
@@ -36,7 +37,7 @@
 #include "dbmanager.h"
 #include "dispatcher.h"
 
-#define COMMIT_INTERVAL_DEFAULT 128
+#define VH_HANDLE dbmanager->valhalla
 
 struct dbmanager_s {
   valhalla_t   *valhalla;
@@ -52,12 +53,20 @@ struct dbmanager_s {
 
   database_t   *database;
   unsigned int  commit_int;
+
+  vh_stats_cnt_t *st_insert;
+  vh_stats_cnt_t *st_update;
+  vh_stats_cnt_t *st_delete;
+  vh_stats_cnt_t *st_nochange;
+  vh_stats_cnt_t *st_cleanup;
 };
 
-typedef struct dbmanager_stats_s {
-  int file_insert, file_update, file_nochange;
-  int grab_insert, grab_update;
-} dbmanager_stats_t;
+#define STATS_GROUP     "dbmanager"
+#define STATS_INSERT    "insert"
+#define STATS_UPDATE    "update"
+#define STATS_DELETE    "delete"
+#define STATS_NOCHANGE  "nochange"
+#define STATS_CLEANUP   "cleanup"
 
 
 static inline int
@@ -87,22 +96,18 @@ vh_dbmanager_extmd_free (dbmanager_extmd_t *extmd)
   free (extmd);
 }
 
-#define METADATA_GRABBER_POST             \
-  vh_metadata_free (pdata->meta_grabber); \
-  pdata->meta_grabber = NULL;             \
-  if (pdata->wait)                        \
-    sem_post (&pdata->sem_grabber);
-
 static int
-dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
+dbmanager_queue (dbmanager_t *dbmanager)
 {
   int res;
   int e;
+  int grab = 0;
   void *data = NULL;
   file_data_t *pdata;
 
   do
   {
+    int interval;
     e = ACTION_NO_OPERATION;
     data = NULL;
 
@@ -115,8 +120,10 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
 
     if (e == ACTION_DB_NEXT_LOOP)
     {
-      vh_dispatcher_action_send (dbmanager->valhalla->dispatcher,
+      vh_dispatcher_action_send (VH_HANDLE->dispatcher,
                                  FIFO_QUEUE_PRIORITY_NORMAL, e, NULL);
+
+      vh_stats_dump (VH_HANDLE->stats, NULL);
       goto out;
     }
 
@@ -134,7 +141,8 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
         continue;
 
       vh_database_metadata_insert (dbmanager->database, extmd->path,
-                                   extmd->meta, extmd->data, extmd->group);
+                                   extmd->meta, extmd->data,
+                                   extmd->lang, extmd->group);
       vh_dbmanager_extmd_free (extmd);
       continue;
     }
@@ -147,7 +155,8 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
         continue;
 
       vh_database_metadata_update (dbmanager->database, extmd->path,
-                                   extmd->meta, extmd->data, extmd->ndata);
+                                   extmd->meta, extmd->data,
+                                   extmd->ndata, extmd->lang);
       vh_dbmanager_extmd_free (extmd);
       continue;
     }
@@ -164,12 +173,27 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
       vh_dbmanager_extmd_free (extmd);
       continue;
     }
+
+    case ACTION_DB_EXT_PRIORITY:
+    {
+      dbmanager_extmd_t *extmd = data;
+
+      if (!extmd)
+        continue;
+
+      vh_database_metadata_priority (dbmanager->database, extmd->path,
+                                     extmd->meta, extmd->data, extmd->priority);
+      vh_dbmanager_extmd_free (extmd);
+      continue;
+    }
     }
 
     /* Manage BEGIN / COMMIT transactions */
-    vh_database_step_transaction (dbmanager->database, dbmanager->commit_int,
-                                  stats->file_insert + stats->file_update +
-                                  stats->grab_insert + stats->grab_update);
+    interval  = (int) vh_stats_counter_read (dbmanager->st_insert);
+    interval += (int) vh_stats_counter_read (dbmanager->st_update);
+    interval += grab;
+    vh_database_step_transaction (dbmanager->database,
+                                  dbmanager->commit_int, interval);
 
     pdata = data;
 
@@ -184,40 +208,56 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
 
     /* received from the dispatcher */
     case ACTION_DB_END:
-      vh_database_file_interrupted_clear (dbmanager->database, pdata->file);
+      vh_database_file_interrupted_clear (dbmanager->database,
+                                          pdata->file.path);
       if (pdata->od != OD_TYPE_DEF)
-        vh_event_handler_send (dbmanager->valhalla->event_handler,
-                               pdata->file, VALHALLA_EVENT_ENDED, NULL);
+        vh_event_handler_od_send (VH_HANDLE->event_handler,
+                                  pdata->file.path,
+                                  VALHALLA_EVENTOD_ENDED, NULL);
       break;
 
     /* received from the dispatcher (grabbed data) */
     case ACTION_DB_INSERT_G:
       vh_database_file_grab_insert (dbmanager->database, pdata);
+    case ACTION_DB_UPDATE_G:
+    {
+      int res;
+
+      if (e == ACTION_DB_UPDATE_G)
+        vh_database_file_grab_update (dbmanager->database, pdata);
+
       if (pdata->od != OD_TYPE_DEF)
-        vh_event_handler_send (dbmanager->valhalla->event_handler, pdata->file,
-                               VALHALLA_EVENT_GRABBED, pdata->grabber_name);
-      METADATA_GRABBER_POST
-      stats->grab_insert++;
+        vh_event_handler_od_send (VH_HANDLE->event_handler,
+                                  pdata->file.path,
+                                  VALHALLA_EVENTOD_GRABBED,
+                                  pdata->grabber_name);
+      res = vh_event_handler_md_send (VH_HANDLE->event_handler,
+                                      VALHALLA_EVENTMD_GRABBER,
+                                      pdata->grabber_name, &pdata->file,
+                                      pdata->meta_grabber);
+      if (res)
+        vh_metadata_free (pdata->meta_grabber);
+      pdata->meta_grabber = NULL;
+
+      if (pdata->wait)
+        sem_post (&pdata->sem_grabber);
+
+      grab++;
       continue;
+    }
 
     /* received from the dispatcher (parsed data) */
     case ACTION_DB_UPDATE_P:
-      stats->file_update++;
+      VH_STATS_COUNTER_INC (dbmanager->st_update);
     case ACTION_DB_INSERT_P:
       vh_database_file_data_update (dbmanager->database, pdata);
       if (pdata->od != OD_TYPE_DEF)
-        vh_event_handler_send (dbmanager->valhalla->event_handler,
-                               pdata->file, VALHALLA_EVENT_PARSED, NULL);
-      continue;
-
-    /* received from the dispatcher (grabbed data) */
-    case ACTION_DB_UPDATE_G:
-      vh_database_file_grab_update (dbmanager->database, pdata);
-      if (pdata->od != OD_TYPE_DEF)
-        vh_event_handler_send (dbmanager->valhalla->event_handler, pdata->file,
-                               VALHALLA_EVENT_GRABBED, pdata->grabber_name);
-      METADATA_GRABBER_POST
-      stats->grab_update++;
+        vh_event_handler_od_send (VH_HANDLE->event_handler,
+                                  pdata->file.path,
+                                  VALHALLA_EVENTOD_PARSED, NULL);
+      vh_event_handler_md_send (VH_HANDLE->event_handler,
+                                VALHALLA_EVENTMD_PARSER, NULL,
+                                &pdata->file, pdata->meta_parser);
       continue;
 
     /* received from the scanner */
@@ -225,7 +265,7 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
     {
       int interrup = 0;
       int64_t mtime =
-        vh_database_file_get_mtime (dbmanager->database, pdata->file);
+        vh_database_file_get_mtime (dbmanager->database, pdata->file.path);
       /*
        * File is parsed only if mtime has changed, if the grabbing/downloading
        * was interrupted or if it is unexistant in the database.
@@ -244,55 +284,55 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
       if (mtime >= 0)
       {
         interrup =
-          vh_database_file_get_interrupted (dbmanager->database, pdata->file);
+          vh_database_file_get_interrupted (dbmanager->database,
+                                            pdata->file.path);
         /*
          * Retrieve the list of all grabbers already handled for this file
          * and search if there are files to download since the interruption.
          * But if mtime has changed, the file must be _fully_ updated.
          */
-        if (interrup == 1 && (int64_t) pdata->mtime == mtime)
+        if (interrup == 1 && pdata->file.mtime == mtime)
         {
           vh_database_file_get_grabber (dbmanager->database,
-                                        pdata->file, &pdata->grabber_list);
-          vh_database_file_get_dlcontext (dbmanager->database,
-                                          pdata->file, &pdata->list_downloader);
+                                        pdata->file.path, pdata->grabber_list);
+          vh_database_file_get_dlcontext (dbmanager->database, pdata->file.path,
+                                          &pdata->list_downloader);
         }
         /*
          * Delete all previous associations on the file because the main
          * metadata have changed.
          */
-        else if ((int64_t) pdata->mtime != mtime)
+        else if (pdata->file.mtime != mtime)
         {
-          vh_database_file_data_delete (dbmanager->database, pdata->file);
-          vh_database_file_grab_delete (dbmanager->database, pdata->file);
+          vh_database_file_data_delete (dbmanager->database, pdata->file.path);
+          vh_database_file_grab_delete (dbmanager->database, pdata->file.path);
         }
       }
       else
       {
         vh_database_file_insert (dbmanager->database, pdata);
-        stats->file_insert++;
+        VH_STATS_COUNTER_INC (dbmanager->st_insert);
       }
 
-      if (mtime < 0 || (int64_t) pdata->mtime != mtime || interrup == 1)
+      if (mtime < 0 || pdata->file.mtime != mtime || interrup == 1)
       {
-        vh_dispatcher_action_send (dbmanager->valhalla->dispatcher,
-                                   pdata->priority,
-                                   mtime < 0
-                                   ? ACTION_DB_INSERT_P : ACTION_DB_UPDATE_P,
-                                   pdata);
+        int act = mtime < 0 ? ACTION_DB_INSERT_P : ACTION_DB_UPDATE_P;
+        vh_dispatcher_action_send (VH_HANDLE->dispatcher,
+                                   pdata->priority, act, pdata);
         continue;
       }
 
       if (pdata->od != OD_TYPE_DEF)
-        vh_event_handler_send (dbmanager->valhalla->event_handler,
-                               pdata->file, VALHALLA_EVENT_ENDED, NULL);
-      stats->file_nochange++;
+        vh_event_handler_od_send (VH_HANDLE->event_handler,
+                                  pdata->file.path,
+                                  VALHALLA_EVENTOD_ENDED, NULL);
+      VH_STATS_COUNTER_INC (dbmanager->st_nochange);
     }
     }
 
     /* Must not come from "On-demand" */
     if (pdata->od == OD_TYPE_DEF || pdata->od == OD_TYPE_UPD)
-      vh_scanner_action_send (dbmanager->valhalla->scanner,
+      vh_scanner_action_send (VH_HANDLE->scanner,
                               FIFO_QUEUE_PRIORITY_NORMAL,
                               ACTION_ACKNOWLEDGE, NULL);
     vh_file_data_free (pdata);
@@ -311,14 +351,17 @@ dbmanager_queue (dbmanager_t *dbmanager, dbmanager_stats_t *stats)
 static void *
 dbmanager_thread (void *arg)
 {
-  int rc;
+  int rc, tid, loop = 0;
   const char *file;
   dbmanager_t *dbmanager = arg;
 
   if (!dbmanager)
     pthread_exit (NULL);
 
-  vh_setpriority (dbmanager->priority);
+  tid = vh_setpriority (dbmanager->priority);
+
+  vh_log (VALHALLA_MSG_VERBOSE,
+          "[%s] tid: %i priority: %i", __FUNCTION__, tid, dbmanager->priority);
 
   /*
    * Change files where interrupted__ is -1 to 1, in the case where Valhalla
@@ -328,18 +371,17 @@ dbmanager_thread (void *arg)
 
   do
   {
-    dbmanager_stats_t stats = {
-      0, 0, 0, 0, 0
-    };
     int stats_delete   = 0;
-    int stats_cleanup  = 0;
+    uint64_t stats_update = 0;
     int rst = 0;
 
+    vh_log (VALHALLA_MSG_INFO, "[%s] Begin loop %i", __FUNCTION__, loop);
+
     /* Clear all checked__ files */
     vh_database_file_checked_clear (dbmanager->database);
 
     vh_database_begin_transaction (dbmanager->database);
-    rc = dbmanager_queue (dbmanager, &stats);
+    rc = dbmanager_queue (dbmanager);
     vh_database_end_transaction (dbmanager->database);
 
     /*
@@ -350,9 +392,11 @@ dbmanager_thread (void *arg)
     while ((file =
               vh_database_file_get_checked_clear (dbmanager->database, rst)))
     {
-      if (vh_scanner_path_cmp (dbmanager->valhalla->scanner, file)
-          || vh_scanner_suffix_cmp (dbmanager->valhalla->scanner, file)
-          || access (file, R_OK))
+      if (dbmanager_is_stopped (dbmanager))
+        rst = 1;
+      else if (vh_scanner_path_cmp (VH_HANDLE->scanner, file)
+               || vh_scanner_suffix_cmp (VH_HANDLE->scanner, file)
+               || access (file, R_OK))
       {
         /* Manage BEGIN / COMMIT transactions */
         vh_database_step_transaction (dbmanager->database,
@@ -361,31 +405,44 @@ dbmanager_thread (void *arg)
         vh_database_file_delete (dbmanager->database, file);
         stats_delete++;
       }
-
-      if (dbmanager_is_stopped (dbmanager))
-        rst = 1;
     }
 
+    /*
+     * Get all files where outofpath__ is set to 1 and check if the file is
+     * valid. The entry is deleted otherwise.
+     */
+    rst = 0;
+    while ((file =
+              vh_database_file_get_outofpath_set (dbmanager->database, rst)))
+    {
+      if (dbmanager_is_stopped (dbmanager))
+        rst = 1;
+      else if (vh_scanner_suffix_cmp (VH_HANDLE->scanner, file)
+               || access (file, R_OK))
+      {
+        /* Manage BEGIN / COMMIT transactions */
+        vh_database_step_transaction (dbmanager->database,
+                                      dbmanager->commit_int, stats_delete);
+
+        vh_database_file_delete (dbmanager->database, file);
+        stats_delete++;
+      }
+    }
+
+    VH_STATS_COUNTER_ACC (dbmanager->st_delete, (unsigned) stats_delete);
+
     /* Clean all relations */
-    if (stats.file_update || stats.grab_update || stats_delete)
-      stats_cleanup = vh_database_cleanup (dbmanager->database);
+    stats_update = vh_stats_counter_read (dbmanager->st_update);
+    if (stats_update || stats_delete)
+    {
+      int val = vh_database_cleanup (dbmanager->database);
+      if (val > 0)
+        VH_STATS_COUNTER_ACC (dbmanager->st_cleanup, (uint64_t) val);
+    }
 
     vh_database_end_transaction (dbmanager->database);
 
-    /* Statistics */
-    if (vh_log_test (VALHALLA_MSG_INFO))
-    {
-      valhalla_log (VALHALLA_MSG_INFO, "[%s] Files inserted    : %i",
-                    __FUNCTION__, stats.file_insert);
-      valhalla_log (VALHALLA_MSG_INFO, "[%s] Files updated     : %i",
-                    __FUNCTION__, stats.file_update);
-      valhalla_log (VALHALLA_MSG_INFO, "[%s] Files deleted     : %i",
-                    __FUNCTION__, stats_delete);
-      valhalla_log (VALHALLA_MSG_INFO, "[%s] Files unchanged   : %i",
-                    __FUNCTION__, stats.file_nochange);
-      valhalla_log (VALHALLA_MSG_INFO, "[%s] Relations cleanup : %i",
-                    __FUNCTION__, stats_cleanup);
-    }
+    vh_log (VALHALLA_MSG_INFO, "[%s] End loop %i", __FUNCTION__, loop++);
   }
   while (rc == ACTION_DB_NEXT_LOOP && !dbmanager_is_stopped (dbmanager));
 
@@ -398,7 +455,7 @@ vh_dbmanager_run (dbmanager_t *dbmanager, int priority)
   int res = DBMANAGER_SUCCESS;
   pthread_attr_t attr;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return DBMANAGER_ERROR_HANDLER;
@@ -423,7 +480,7 @@ vh_dbmanager_run (dbmanager_t *dbmanager, int priority)
 fifo_queue_t *
 vh_dbmanager_fifo_get (dbmanager_t *dbmanager)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return NULL;
@@ -434,7 +491,7 @@ vh_dbmanager_fifo_get (dbmanager_t *dbmanager)
 void
 vh_dbmanager_pause (dbmanager_t *dbmanager)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -445,7 +502,7 @@ vh_dbmanager_pause (dbmanager_t *dbmanager)
 void
 vh_dbmanager_wait (dbmanager_t *dbmanager)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -460,7 +517,7 @@ vh_dbmanager_wait (dbmanager_t *dbmanager)
 void
 vh_dbmanager_stop (dbmanager_t *dbmanager, int f)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -474,6 +531,8 @@ vh_dbmanager_stop (dbmanager_t *dbmanager, int f)
     vh_fifo_queue_push (dbmanager->fifo,
                         FIFO_QUEUE_PRIORITY_HIGH, ACTION_KILL_THREAD, NULL);
     dbmanager->wait = 1;
+
+    VH_THREAD_PAUSE_FORCESTOP (dbmanager, 1)
   }
 
   if (f & STOP_FLAG_WAIT && dbmanager->wait)
@@ -486,7 +545,7 @@ vh_dbmanager_stop (dbmanager_t *dbmanager, int f)
 void
 vh_dbmanager_uninit (dbmanager_t *dbmanager)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -501,12 +560,36 @@ vh_dbmanager_uninit (dbmanager_t *dbmanager)
   free (dbmanager);
 }
 
+static void
+dbmanager_stats_dump (vh_stats_t *stats, void *data)
+{
+  dbmanager_t *dbmanager = data;
+
+  if (!stats || !dbmanager)
+    return;
+
+  vh_log (VALHALLA_MSG_INFO, "==============================");
+  vh_log (VALHALLA_MSG_INFO, "Statistics dump (" STATS_GROUP ")");
+  vh_log (VALHALLA_MSG_INFO, "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
+
+  vh_log (VALHALLA_MSG_INFO, "Files inserted    | %"PRIu64,
+          vh_stats_counter_read (dbmanager->st_insert));
+  vh_log (VALHALLA_MSG_INFO, "Files updated     | %"PRIu64,
+          vh_stats_counter_read (dbmanager->st_update));
+  vh_log (VALHALLA_MSG_INFO, "Files deleted     | %"PRIu64,
+          vh_stats_counter_read (dbmanager->st_delete));
+  vh_log (VALHALLA_MSG_INFO, "Files unchanged   | %"PRIu64,
+          vh_stats_counter_read (dbmanager->st_nochange));
+  vh_log (VALHALLA_MSG_INFO, "Relations cleaned | %"PRIu64,
+          vh_stats_counter_read (dbmanager->st_cleanup));
+}
+
 dbmanager_t *
 vh_dbmanager_init (valhalla_t *handle, const char *db, unsigned int commit_int)
 {
   dbmanager_t *dbmanager;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle)
     return NULL;
@@ -523,15 +606,29 @@ vh_dbmanager_init (valhalla_t *handle, const char *db, unsigned int commit_int)
   if (!dbmanager->database)
     goto err;
 
-  if (commit_int <= 0)
-    commit_int = COMMIT_INTERVAL_DEFAULT;
+  if (!commit_int)
+    commit_int = DBMANAGER_COMMIT_INTERVAL_DEF;
   dbmanager->commit_int = commit_int;
 
-  dbmanager->valhalla = handle;
+  dbmanager->valhalla = handle; /* VH_HANDLE */
 
   pthread_mutex_init (&dbmanager->mutex_run, NULL);
   VH_THREAD_PAUSE_INIT (dbmanager)
 
+  /* init statistics */
+  vh_stats_grp_add (handle->stats,
+                    STATS_GROUP, dbmanager_stats_dump, dbmanager);
+  dbmanager->st_insert =
+    vh_stats_grp_counter_add (handle->stats, STATS_GROUP, STATS_INSERT,   NULL);
+  dbmanager->st_update =
+    vh_stats_grp_counter_add (handle->stats, STATS_GROUP, STATS_UPDATE,   NULL);
+  dbmanager->st_delete =
+    vh_stats_grp_counter_add (handle->stats, STATS_GROUP, STATS_DELETE,   NULL);
+  dbmanager->st_nochange =
+    vh_stats_grp_counter_add (handle->stats, STATS_GROUP, STATS_NOCHANGE, NULL);
+  dbmanager->st_cleanup =
+    vh_stats_grp_counter_add (handle->stats, STATS_GROUP, STATS_CLEANUP,  NULL);
+
   return dbmanager;
 
  err:
@@ -543,7 +640,7 @@ void
 vh_dbmanager_action_send (dbmanager_t *dbmanager,
                           fifo_queue_prio_t prio, int action, void *data)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -557,7 +654,7 @@ vh_dbmanager_file_complete (dbmanager_t *dbmanager,
 {
   int res;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager || !file)
     return 0;
@@ -569,8 +666,8 @@ vh_dbmanager_file_complete (dbmanager_t *dbmanager,
     if (mt != mtime)
       res = 1; /* must be updated */
     else
-      vh_event_handler_send (dbmanager->valhalla->event_handler,
-                             file, VALHALLA_EVENT_ENDED, NULL);
+      vh_event_handler_od_send (VH_HANDLE->event_handler,
+                                file, VALHALLA_EVENTOD_ENDED, NULL);
   }
 
   return !res;
@@ -579,7 +676,7 @@ vh_dbmanager_file_complete (dbmanager_t *dbmanager,
 void
 vh_dbmanager_db_dlcontext_save (dbmanager_t *dbmanager, file_data_t *data)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager || !data)
     return;
@@ -590,7 +687,7 @@ vh_dbmanager_db_dlcontext_save (dbmanager_t *dbmanager, file_data_t *data)
 void
 vh_dbmanager_db_dlcontext_delete (dbmanager_t *dbmanager)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -601,7 +698,7 @@ vh_dbmanager_db_dlcontext_delete (dbmanager_t *dbmanager)
 void
 vh_dbmanager_db_begin_transaction (dbmanager_t *dbmanager)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -612,7 +709,7 @@ vh_dbmanager_db_begin_transaction (dbmanager_t *dbmanager)
 void
 vh_dbmanager_db_end_transaction (dbmanager_t *dbmanager)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
     return;
@@ -620,50 +717,78 @@ vh_dbmanager_db_end_transaction (dbmanager_t *dbmanager)
   vh_database_end_transaction (dbmanager->database);
 }
 
-int
+valhalla_db_stmt_t *
 vh_dbmanager_db_metalist_get (dbmanager_t *dbmanager,
                               valhalla_db_item_t *search,
-                              valhalla_db_restrict_t *restriction,
-                              int (*select_cb) (void *data,
-                                                valhalla_db_metares_t *res),
-                              void *data)
+                              valhalla_file_type_t filetype,
+                              valhalla_db_restrict_t *restriction)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
-    return -1;
+    return NULL;
 
   return vh_database_metalist_get (dbmanager->database,
-                                   search, restriction, select_cb, data);
+                                   search, filetype, restriction);
 }
 
-int
+const valhalla_db_metares_t *
+vh_dbmanager_db_metalist_read (dbmanager_t *dbmanager,
+                               valhalla_db_stmt_t *vhstmt)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!dbmanager)
+    return NULL;
+
+  return vh_database_metalist_read (dbmanager->database, vhstmt);
+}
+
+valhalla_db_stmt_t *
 vh_dbmanager_db_filelist_get (dbmanager_t *dbmanager,
                               valhalla_file_type_t filetype,
-                              valhalla_db_restrict_t *restriction,
-                              int (*select_cb) (void *data,
-                                                valhalla_db_fileres_t *res),
-                              void *data)
+                              valhalla_db_restrict_t *restriction)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
-    return -1;
+    return NULL;
 
-  return vh_database_filelist_get (dbmanager->database,
-                                   filetype, restriction, select_cb, data);
+  return vh_database_filelist_get (dbmanager->database, filetype, restriction);
 }
 
-int
+const valhalla_db_fileres_t *
+vh_dbmanager_db_filelist_read (dbmanager_t *dbmanager,
+                               valhalla_db_stmt_t *vhstmt)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!dbmanager)
+    return NULL;
+
+  return vh_database_filelist_read (dbmanager->database, vhstmt);
+}
+
+valhalla_db_stmt_t *
 vh_dbmanager_db_file_get (dbmanager_t *dbmanager,
                           int64_t id, const char *path,
-                          valhalla_db_restrict_t *restriction,
-                          valhalla_db_filemeta_t **res)
+                          valhalla_db_restrict_t *restriction)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!dbmanager)
-    return -1;
+    return NULL;
 
-  return vh_database_file_get (dbmanager->database, id, path, restriction, res);
+  return vh_database_file_get (dbmanager->database, id, path, restriction);
+}
+
+const valhalla_db_metares_t *
+vh_dbmanager_db_file_read (dbmanager_t *dbmanager, valhalla_db_stmt_t *vhstmt)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!dbmanager)
+    return NULL;
+
+  return vh_database_file_read (dbmanager->database, vhstmt);
 }
