@@ -23,6 +23,6 @@
 #ifndef VALHALLA_GRABBER_IMDB_H
 #define VALHALLA_GRABBER_IMDB_H
 
-grabber_list_t *vh_grabber_imdb_register (void);
+grabber_list_t *vh_grabber_imdb_register (struct url_ctl_s *url_ctl);
 
 #endif /* VALHALLA_GRABBER_IMDB_H */
