@@ -30,13 +30,12 @@
 #include "xml_utils.h"
 #include "url_utils.h"
 #include "utils.h"
+#include "osdep.h"
 #include "logs.h"
 
 #define GRABBER_CAP_FLAGS \
   GRABBER_CAP_AUDIO
 
-#define MAX_URL_SIZE      1024
-
 #define LYRICWIKI_HOSTNAME     "lyrics.wikia.com"
 #define LYRICWIKI_QUERY_SEARCH "http://%s/api.php?func=getSong&artist=%s&song=%s&fmt=xml"
 
@@ -50,11 +49,16 @@ static const char *lyricwiki_offset[][2] = {
 
 typedef struct grabber_lyricwiki_s {
   url_t *handler;
+  const metadata_plist_t *pl;
 } grabber_lyricwiki_t;
 
+static const metadata_plist_t lyricwiki_pl[] = {
+  { NULL,                             VALHALLA_METADATA_PL_NORMAL   }
+};
+
 
 static int
-grabber_lyricwiki_get (url_t *handler, file_data_t *fdata,
+grabber_lyricwiki_get (grabber_lyricwiki_t *lyricwiki, file_data_t *fdata,
                        const char *artist, const char *song)
 {
   char url[MAX_URL_SIZE];
@@ -64,20 +68,19 @@ grabber_lyricwiki_get (url_t *handler, file_data_t *fdata,
   xmlDocPtr doc;
 
   /* proceed with LyricWiki.org search request */
-  memset (url, '\0', MAX_URL_SIZE);
-  snprintf (url, MAX_URL_SIZE, LYRICWIKI_QUERY_SEARCH,
+  snprintf (url, sizeof (url), LYRICWIKI_QUERY_SEARCH,
             LYRICWIKI_HOSTNAME, artist, song);
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, "Search Request: %s", url);
+  vh_log (VALHALLA_MSG_VERBOSE, "Search Request: %s", url);
 
-  udata = vh_url_get_data (handler, url);
+  udata = vh_url_get_data (lyricwiki->handler, url);
   if (udata.status != 0)
     return -1;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, "Search Reply: %s", udata.buffer);
+  vh_log (VALHALLA_MSG_VERBOSE, "Search Reply: %s", udata.buffer);
 
   /* parse the XML answer */
-  doc = vh_get_xml_doc_from_memory (udata.buffer);
+  doc = vh_xml_get_doc_from_memory (udata.buffer);
   free (udata.buffer);
 
   if (!doc)
@@ -95,7 +98,7 @@ grabber_lyricwiki_get (url_t *handler, file_data_t *fdata,
     unsigned int i, j;
     int cmp;
 
-    udata = vh_url_get_data (handler, html);
+    udata = vh_url_get_data (lyricwiki->handler, html);
     free (html);
     if (udata.status != 0)
       return -1;
@@ -160,8 +163,8 @@ grabber_lyricwiki_get (url_t *handler, file_data_t *fdata,
       j++;
     }
 
-    vh_metadata_add_auto (&fdata->meta_grabber,
-                          VALHALLA_METADATA_LYRICS, lyrics);
+    vh_metadata_add_auto (&fdata->meta_grabber, VALHALLA_METADATA_LYRICS,
+                          lyrics, VALHALLA_LANG_EN, lyricwiki->pl);
 
     free (txt);
     free (lyrics);
@@ -178,22 +181,23 @@ grabber_lyricwiki_get (url_t *handler, file_data_t *fdata,
 static void *
 grabber_lyricwiki_priv (void)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   return calloc (1, sizeof (grabber_lyricwiki_t));
 }
 
 static int
-grabber_lyricwiki_init (void *priv)
+grabber_lyricwiki_init (void *priv, const grabber_param_t *param)
 {
   grabber_lyricwiki_t *lyricwiki = priv;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!lyricwiki)
     return -1;
 
-  lyricwiki->handler = vh_url_new ();
+  lyricwiki->handler = vh_url_new (param->url_ctl);
+  lyricwiki->pl      = param->pl;
   return lyricwiki->handler ? 0 : -1;
 }
 
@@ -202,7 +206,7 @@ grabber_lyricwiki_uninit (void *priv)
 {
   grabber_lyricwiki_t *lyricwiki = priv;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!lyricwiki)
     return;
@@ -215,11 +219,11 @@ static int
 grabber_lyricwiki_grab (void *priv, file_data_t *data)
 {
   grabber_lyricwiki_t *lyricwiki = priv;
-  metadata_t *title = NULL, *author = NULL;
+  const metadata_t *title = NULL, *author = NULL;
   char *artist, *song;
   int res, err;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   err = vh_metadata_get (data->meta_parser, "title", 0, &title);
   if (err)
@@ -243,7 +247,7 @@ grabber_lyricwiki_grab (void *priv, file_data_t *data)
   artist = vh_url_escape_string (lyricwiki->handler, author->value);
   song = vh_url_escape_string (lyricwiki->handler, title->value);
 
-  res = grabber_lyricwiki_get (lyricwiki->handler, data, artist, song);
+  res = grabber_lyricwiki_get (lyricwiki, data, artist, song);
 
   free (artist);
   free (song);
@@ -258,6 +262,8 @@ grabber_lyricwiki_grab (void *priv, file_data_t *data)
 /* vh_grabber_lyricwiki_register () */
 GRABBER_REGISTER (lyricwiki,
                   GRABBER_CAP_FLAGS,
+                  lyricwiki_pl,
+                  0,
                   grabber_lyricwiki_priv,
                   grabber_lyricwiki_init,
                   grabber_lyricwiki_uninit,
