@@ -30,12 +30,19 @@
 #include "url_utils.h"
 #include "grabber_utils.h"
 #include "utils.h"
+#include "osdep.h"
 #include "logs.h"
 #include "md5.h"
 
 #define GRABBER_CAP_FLAGS \
   GRABBER_CAP_VIDEO
 
+/*
+ * The documentation is available on:
+ *  http://thetvdb.com/wiki/index.php?title=Programmers_API
+ *  http://www.thetvdb.com/api/
+ */
+
 /*
  * When defined, an API unofficial (undocumented) is used in order to found
  * a name which can be used to return something interesting in the result,
@@ -43,8 +50,6 @@
  */
 #define GRABBER_TVDB_UNOFFICIAL_API
 
-#define MAX_URL_SIZE      1024
-
 #define TVDB_DEFAULT_LANGUAGE   "en"
 
 #define TVDB_HOSTNAME           "thetvdb.com"
@@ -56,61 +61,164 @@
 #define TVDB_QUERY_SEARCH       "http://%s/api/GetSeries.php?seriesname=%s"
 #define TVDB_QUERY_SEARCH_NEW   "http://%s/api/GetSeriesNew.php?seriesname=%s"
 #define TVDB_QUERY_INFO         "http://%s/api/%s/series/%s/%s.xml"
+#define TVDB_EPISODE_INFO       "http://%s/api/%s/series/%s/default/%u/%u/%s.xml"
 #define TVDB_COVERS_URL         "http://%s/banners/%s"
 
 typedef struct grabber_tvdb_s {
   url_t *handler;
+  const metadata_plist_t *pl;
 } grabber_tvdb_t;
 
+static const metadata_plist_t tvdb_pl[] = {
+  { VALHALLA_METADATA_COVER,          VALHALLA_METADATA_PL_ABOVE    },
+  { VALHALLA_METADATA_FAN_ART,        VALHALLA_METADATA_PL_ABOVE    },
+  { NULL,                             VALHALLA_METADATA_PL_NORMAL   }
+};
+
 
 static void
-grabber_tvdb_parse_genre (file_data_t *fdata, xmlChar *genre)
+grabber_tvdb_parse_list (file_data_t *fdata, xmlNode *node, const char *tag,
+                         const char *name, valhalla_lang_t lang,
+                         const metadata_plist_t *pl)
 {
-  char *category, *saveptr = NULL;
+  char *value, *saveptr = NULL;
+  xmlChar *tmp = NULL;
 
-  if (!fdata || !genre)
+  if (!fdata || !node || !tag || !name)
     return;
 
-  category = strtok_r ((char *) genre, "|", &saveptr);
-  while (category)
+  tmp = vh_xml_get_prop_value_from_tree (node, tag);
+  if (tmp)
   {
-    vh_metadata_add_auto (&fdata->meta_grabber,
-                          VALHALLA_METADATA_CATEGORY, category);
-    category = strtok_r (NULL, "|", &saveptr);
+    value = strtok_r ((char *) tmp, "|", &saveptr);
+    while (value)
+    {
+      vh_metadata_add_auto (&fdata->meta_grabber, name, value, lang, pl);
+      value = strtok_r (NULL, "|", &saveptr);
+    }
+    xmlFree (tmp);
   }
 }
 
 static void
 grabber_tvdb_get_picture (file_data_t *fdata, const char *keywords,
-                          xmlChar *url, valhalla_dl_t dl)
+                          xmlChar *url, const char *metadata_name,
+                          const metadata_plist_t *pl)
 {
   char name[1024] = { 0 };
   char complete_url[MAX_URL_SIZE] = { 0 };
-  char *type, *cover = NULL;
+  char *cover = NULL;
+  valhalla_dl_t dl = VALHALLA_DL_COVER;
 
   if (!fdata || !url)
     return;
 
-  if (dl == VALHALLA_DL_COVER)
-    type = VALHALLA_METADATA_COVER;
-  else if (dl == VALHALLA_DL_FAN_ART)
-    type = VALHALLA_METADATA_FAN_ART;
-  else
-    return;
+  if (!strcmp (metadata_name, VALHALLA_METADATA_FAN_ART))
+    dl = VALHALLA_DL_FAN_ART;
 
   /* constructing url for download tvdb covers and fan art */
-  memset (complete_url, '\0', MAX_URL_SIZE);
-  snprintf (complete_url, MAX_URL_SIZE, TVDB_COVERS_URL,
+  snprintf (complete_url, sizeof (complete_url), TVDB_COVERS_URL,
             TVDB_IMAGES_HOSTNAME, (char *) url);
 
-  snprintf (name, sizeof (name), "%s-%s", type, keywords);
+  snprintf (name, sizeof (name), "%s-%s", metadata_name, keywords);
   cover = vh_md5sum (name);
 
-  vh_metadata_add_auto (&fdata->meta_grabber, type, cover);
+  vh_metadata_add_auto (&fdata->meta_grabber,
+                        metadata_name, cover, VALHALLA_LANG_UNDEF, pl);
   vh_file_dl_add (&fdata->list_downloader, complete_url, cover, dl);
   free (cover);
 }
 
+static int
+grabber_tvdb_get_episode (grabber_tvdb_t *tvdb, file_data_t *fdata,
+                          const char *keywords, const char *seriesid)
+{
+  char name[1024] = { 0 };
+  const metadata_t *tag = NULL;
+  int err;
+  unsigned int season, episode;
+  char url[MAX_URL_SIZE];
+  url_data_t udata;
+  xmlDocPtr doc = NULL;
+  xmlChar *tmp = NULL;
+  xmlNode *n;
+
+  err = vh_metadata_get (fdata->meta_parser,
+                         VALHALLA_METADATA_SEASON, 0, &tag);
+  if (err)
+    return -1;
+  season = atoi (tag->value);
+
+  err = vh_metadata_get (fdata->meta_parser,
+                         VALHALLA_METADATA_EPISODE, 0, &tag);
+  if (err)
+    return -1;
+  episode = atoi (tag->value);
+
+  /* proceed with TVDB episode request */
+  snprintf (url, sizeof (url), TVDB_EPISODE_INFO,
+            TVDB_HOSTNAME, TVDB_API_KEY, seriesid, season, episode,
+            TVDB_DEFAULT_LANGUAGE);
+
+  vh_log (VALHALLA_MSG_VERBOSE, "Episode Info Request: %s", url);
+
+  udata = vh_url_get_data (tvdb->handler, url);
+  if (udata.status != 0)
+    goto error;
+
+  vh_log (VALHALLA_MSG_VERBOSE, "Episode Info Reply: %s", udata.buffer);
+
+  /* parse the XML answer */
+  doc = vh_xml_get_doc_from_memory (udata.buffer);
+  free (udata.buffer);
+  if (!doc)
+    goto error;
+
+  n = xmlDocGetRootElement (doc);
+
+  /* fetch tv show overview description */
+  vh_grabber_parse_str (fdata, n, "Overview", VALHALLA_METADATA_SYNOPSIS_SHOW,
+                        VALHALLA_LANG_EN, tvdb->pl);
+
+  /* fetch tv show first air date */
+  vh_grabber_parse_str (fdata, n, "FirstAired", VALHALLA_METADATA_PREMIERED,
+                        VALHALLA_LANG_UNDEF, tvdb->pl);
+
+  /* fetch tv show directors */
+  grabber_tvdb_parse_list (fdata, n, "Director", VALHALLA_METADATA_DIRECTOR,
+                           VALHALLA_LANG_UNDEF, tvdb->pl);
+
+  /* fetch tv show guest stars */
+  grabber_tvdb_parse_list (fdata, n, "GuestStars", VALHALLA_METADATA_ACTOR,
+                           VALHALLA_LANG_UNDEF, tvdb->pl);
+
+  /* fetch tv show writers */
+  grabber_tvdb_parse_list (fdata, n, "Writer", VALHALLA_METADATA_WRITER,
+                           VALHALLA_LANG_UNDEF, tvdb->pl);
+
+  /* fetch tv show poster */
+  tmp = vh_xml_get_prop_value_from_tree (n, "filename");
+  if (tmp)
+  {
+    if (*tmp)
+    {
+      snprintf (name, sizeof (name), "%s-%d-%d", keywords, season, episode);
+      grabber_tvdb_get_picture (fdata, name, tmp,
+                                VALHALLA_METADATA_COVER_SHOW, tvdb->pl);
+    }
+    xmlFree (tmp);
+  }
+
+  xmlFreeDoc (doc);
+  return 0;
+
+ error:
+  if (doc)
+    xmlFreeDoc (doc);
+
+  return -1;
+}
+
 static char *
 grabber_tvdb_search (url_t *handler, const char *escaped_keywords,
                      const char *query, const char *item, const char *value)
@@ -124,35 +232,34 @@ grabber_tvdb_search (url_t *handler, const char *escaped_keywords,
   xmlNode *n;
 
   /* proceed with TVDB search request */
-  memset (url, '\0', MAX_URL_SIZE);
-  snprintf (url, MAX_URL_SIZE, query, TVDB_HOSTNAME, escaped_keywords);
+  snprintf (url, sizeof (url), query, TVDB_HOSTNAME, escaped_keywords);
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, "Search Request: %s", url);
+  vh_log (VALHALLA_MSG_VERBOSE, "Search Request: %s", url);
 
   udata = vh_url_get_data (handler, url);
   if (udata.status != 0)
     return NULL;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, "Search Reply: %s", udata.buffer);
+  vh_log (VALHALLA_MSG_VERBOSE, "Search Reply: %s", udata.buffer);
 
   /* parse the XML answer */
-  doc = vh_get_xml_doc_from_memory (udata.buffer);
+  doc = vh_xml_get_doc_from_memory (udata.buffer);
   free (udata.buffer);
 
   if (!doc)
     return NULL;
 
   /* check for a known DB entry */
-  n = vh_get_node_xml_tree (xmlDocGetRootElement (doc), item);
+  n = vh_xml_get_node_tree (xmlDocGetRootElement (doc), item);
   if (!n)
   {
-    valhalla_log (VALHALLA_MSG_VERBOSE,
-                  "Unable to find the item \"%s\"", escaped_keywords);
+    vh_log (VALHALLA_MSG_VERBOSE,
+            "Unable to find the item \"%s\"", escaped_keywords);
   }
   else
   {
     /* get value */
-    tmp = vh_get_prop_value_from_xml_tree (n, value);
+    tmp = vh_xml_get_prop_value_from_tree (n, value);
     if (tmp)
     {
       res = strdup ((const char *) tmp);
@@ -165,13 +272,13 @@ grabber_tvdb_search (url_t *handler, const char *escaped_keywords,
 }
 
 static int
-grabber_tvdb_get (url_t *handler, file_data_t *fdata,
+grabber_tvdb_get (grabber_tvdb_t *tvdb, file_data_t *fdata,
                   char *orig_keywords, char *escaped_keywords)
 {
   char url[MAX_URL_SIZE];
   url_data_t udata;
   int res_int = 0;
-  char *title, *seriesid, *keywords;
+  char *title, *seriesid = NULL, *keywords;
 #ifdef GRABBER_TVDB_UNOFFICIAL_API
   char *tmp2;
 #endif /* GRABBER_TVDB_UNOFFICIAL_API */
@@ -187,7 +294,7 @@ grabber_tvdb_get (url_t *handler, file_data_t *fdata,
   (void) orig_keywords;
 
   /* search the exact name for a movie */
-  tmp2 = grabber_tvdb_search (handler, escaped_keywords,
+  tmp2 = grabber_tvdb_search (tvdb->handler, escaped_keywords,
                               TVDB_QUERY_SEARCH_NEW, "Series", "SeriesName");
   if (!tmp2)
     return -1;
@@ -198,7 +305,7 @@ grabber_tvdb_get (url_t *handler, file_data_t *fdata,
     free (tmp2);
     return -1;
   }
-  title = vh_url_escape_string (handler, tmp2);
+  title = vh_url_escape_string (tvdb->handler, tmp2);
   free (tmp2);
 #else
   if (!orig_keywords)
@@ -213,28 +320,26 @@ grabber_tvdb_get (url_t *handler, file_data_t *fdata,
   if (!title)
     goto error;
 
-  seriesid = grabber_tvdb_search (handler, title,
+  seriesid = grabber_tvdb_search (tvdb->handler, title,
                                   TVDB_QUERY_SEARCH, "Series", "seriesid");
   free (title);
   if (!seriesid)
     goto error;
 
   /* proceed with TVDB search request */
-  memset (url, '\0', MAX_URL_SIZE);
-  snprintf (url, MAX_URL_SIZE, TVDB_QUERY_INFO,
+  snprintf (url, sizeof (url), TVDB_QUERY_INFO,
             TVDB_HOSTNAME, TVDB_API_KEY, seriesid, TVDB_DEFAULT_LANGUAGE);
-  free (seriesid);
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, "Info Request: %s", url);
+  vh_log (VALHALLA_MSG_VERBOSE, "Info Request: %s", url);
 
-  udata = vh_url_get_data (handler, url);
+  udata = vh_url_get_data (tvdb->handler, url);
   if (udata.status != 0)
     goto error;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, "Info Reply: %s", udata.buffer);
+  vh_log (VALHALLA_MSG_VERBOSE, "Info Reply: %s", udata.buffer);
 
   /* parse the XML answer */
-  doc = vh_get_xml_doc_from_memory (udata.buffer);
+  doc = vh_xml_get_doc_from_memory (udata.buffer);
   free (udata.buffer);
   if (!doc)
     goto error;
@@ -242,52 +347,67 @@ grabber_tvdb_get (url_t *handler, file_data_t *fdata,
   n = xmlDocGetRootElement (doc);
 
   /* fetch tv show overview description */
-  vh_grabber_parse_str (fdata, n, "Overview", VALHALLA_METADATA_SYNOPSIS);
+  vh_grabber_parse_str (fdata, n, "Overview",
+                        VALHALLA_METADATA_SYNOPSIS, VALHALLA_LANG_EN, tvdb->pl);
 
   /* fetch tv show first air date */
   vh_xml_search_year (n, "FirstAired", &res_int);
   if (res_int)
   {
-    vh_grabber_parse_int (fdata, res_int, VALHALLA_METADATA_YEAR);
+    vh_grabber_parse_int (fdata, res_int, VALHALLA_METADATA_YEAR, tvdb->pl);
     res_int = 0;
   }
 
   /* fetch tv show categories */
-  tmp = vh_get_prop_value_from_xml_tree (n, "Genre");
-  if (tmp)
-  {
-    grabber_tvdb_parse_genre (fdata, tmp);
-    xmlFree(tmp);
-  }
+  grabber_tvdb_parse_list (fdata, n, "Genre", VALHALLA_METADATA_CATEGORY,
+                           VALHALLA_LANG_EN, tvdb->pl);
+
+  /* fetch tv show actors */
+  grabber_tvdb_parse_list (fdata, n, "Actors", VALHALLA_METADATA_ACTOR,
+                           VALHALLA_LANG_UNDEF, tvdb->pl);
 
   /* fetch tv show runtime (in minutes) */
-  vh_grabber_parse_str (fdata, n, "Runtime", VALHALLA_METADATA_RUNTIME);
+  vh_grabber_parse_str (fdata, n, "Runtime", VALHALLA_METADATA_RUNTIME,
+                        VALHALLA_LANG_UNDEF, tvdb->pl);
+
+  /* fetch tv show content rating */
+  vh_grabber_parse_str (fdata, n, "ContentRating", VALHALLA_METADATA_MPAA,
+                        VALHALLA_LANG_UNDEF, tvdb->pl);
 
   /* fetch tv show poster */
-  tmp = vh_get_prop_value_from_xml_tree (n, "poster");
-  if (tmp && *tmp)
+  tmp = vh_xml_get_prop_value_from_tree (n, "poster");
+  if (tmp)
   {
-    grabber_tvdb_get_picture (fdata, keywords, tmp, VALHALLA_DL_COVER);
+    if (*tmp)
+      grabber_tvdb_get_picture (fdata, keywords, tmp,
+                                VALHALLA_METADATA_COVER, tvdb->pl);
     xmlFree (tmp);
   }
 
   /* fetch tv show fan art */
-  tmp = vh_get_prop_value_from_xml_tree (n, "fanart");
-  if (tmp && *tmp)
+  tmp = vh_xml_get_prop_value_from_tree (n, "fanart");
+  if (tmp)
   {
-    grabber_tvdb_get_picture (fdata, keywords, tmp, VALHALLA_DL_FAN_ART);
+    if (*tmp)
+      grabber_tvdb_get_picture (fdata, keywords, tmp,
+                                VALHALLA_METADATA_FAN_ART, tvdb->pl);
     xmlFree (tmp);
   }
 
+  grabber_tvdb_get_episode (tvdb, fdata, keywords, seriesid);
+
   xmlFreeDoc (doc);
   free (keywords);
+  if (seriesid)
+    free (seriesid);
   return 0;
 
  error:
   if (doc)
     xmlFreeDoc (doc);
   free (keywords);
-
+  if (seriesid)
+    free (seriesid);
   return -1;
 }
 
@@ -298,22 +418,23 @@ grabber_tvdb_get (url_t *handler, file_data_t *fdata,
 static void *
 grabber_tvdb_priv (void)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   return calloc (1, sizeof (grabber_tvdb_t));
 }
 
 static int
-grabber_tvdb_init (void *priv)
+grabber_tvdb_init (void *priv, const grabber_param_t *param)
 {
   grabber_tvdb_t *tvdb = priv;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!tvdb)
     return -1;
 
-  tvdb->handler = vh_url_new ();
+  tvdb->handler = vh_url_new (param->url_ctl);
+  tvdb->pl      = param->pl;
   return tvdb->handler ? 0 : -1;
 }
 
@@ -322,7 +443,7 @@ grabber_tvdb_uninit (void *priv)
 {
   grabber_tvdb_t *tvdb = priv;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!tvdb)
     return;
@@ -335,11 +456,11 @@ static int
 grabber_tvdb_grab (void *priv, file_data_t *data)
 {
   grabber_tvdb_t *tvdb = priv;
-  metadata_t *tag = NULL;
+  const metadata_t *tag = NULL;
   char *keywords;
   int err;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   err = vh_metadata_get (data->meta_parser, "title", 0, &tag);
   if (err)
@@ -350,7 +471,7 @@ grabber_tvdb_grab (void *priv, file_data_t *data)
   if (!keywords)
     return -2;
 
-  err = grabber_tvdb_get (tvdb->handler, data, tag->value, keywords);
+  err = grabber_tvdb_get (tvdb, data, tag->value, keywords);
   free (keywords);
 
   return err;
@@ -363,6 +484,8 @@ grabber_tvdb_grab (void *priv, file_data_t *data)
 /* vh_grabber_tvdb_register () */
 GRABBER_REGISTER (tvdb,
                   GRABBER_CAP_FLAGS,
+                  tvdb_pl,
+                  0,
                   grabber_tvdb_priv,
                   grabber_tvdb_init,
                   grabber_tvdb_uninit,
