@@ -25,17 +25,11 @@
 typedef struct list_s list_t;
 
 
-void vh_list_append (list_t **list, const void *data, size_t len);
-void vh_list_free (list_t *list, void (*free_fct) (void *data));
-void *vh_list_search (list_t *list, const void *tocmp,
+void vh_list_append (list_t *list, const void *data, size_t len);
+list_t *vh_list_new (unsigned int depth, void (*free_fct) (void *data));
+void vh_list_free (list_t *list);
+void vh_list_empty (list_t *list);
+void *vh_list_search (const list_t *list, const void *tocmp,
                       int (*cmp_fct) (const void *tocmp, const void *data));
 
-#define VH_LIST_FREE(list, free_fct) \
-  do                                 \
-  {                                  \
-    vh_list_free (list, free_fct);   \
-    list = NULL;                     \
-  }                                  \
-  while (0)
-
 #endif /* VALHALLA_LIST_H */
