@@ -27,8 +27,10 @@
 #include "valhalla.h"
 #include "valhalla_internals.h"
 #include "utils.h"
+#include "osdep.h"
 #include "fifo_queue.h"
 #include "logs.h"
+#include "stats.h"
 #include "thread_utils.h"
 #include "dbmanager.h"
 #include "dispatcher.h"
@@ -41,6 +43,8 @@
 #include "downloader.h"
 #endif /* USE_GRABBER */
 
+#define VH_HANDLE ondemand->valhalla
+
 struct ondemand_s {
   valhalla_t   *valhalla;
   pthread_t     thread;
@@ -50,8 +54,13 @@ struct ondemand_s {
   int             wait;
   int             run;
   pthread_mutex_t mutex_run;
+
+  vh_stats_cnt_t *st_cnt;
+  vh_stats_tmr_t *st_tmr;
 };
 
+#define STATS_GROUP "ondemand"
+
 
 static inline int
 ondemand_is_stopped (ondemand_t *ondemand)
@@ -80,7 +89,7 @@ ondemand_cmp_fct (const void *tocmp, int id, const void *data)
   case ACTION_DB_UPDATE_G:
   case ACTION_DB_END:
   case ACTION_DB_NEWFILE:
-    return strcmp (file, fdata->file);
+    return strcmp (file, fdata->file.path);
 
   default:
     return -1;
@@ -90,7 +99,7 @@ ondemand_cmp_fct (const void *tocmp, int id, const void *data)
 static void *
 ondemand_thread (void *arg)
 {
-  int res;
+  int res, tid;
   int e;
   unsigned int i = 0;
   void *data = NULL;
@@ -120,15 +129,18 @@ ondemand_thread (void *arg)
   if (!ondemand)
     pthread_exit (NULL);
 
-  vh_setpriority (ondemand->priority);
+  tid = vh_setpriority (ondemand->priority);
+
+  vh_log (VALHALLA_MSG_VERBOSE,
+          "[%s] tid: %i priority: %i", __FUNCTION__, tid, ondemand->priority);
 
 #ifdef USE_GRABBER
-  pause[i++].handler = ondemand->valhalla->grabber;
-  pause[i++].handler = ondemand->valhalla->downloader;
+  pause[i++].handler = VH_HANDLE->grabber;
+  pause[i++].handler = VH_HANDLE->downloader;
 #endif /* USE_GRABBER */
-  pause[i++].handler = ondemand->valhalla->parser;
-  pause[i++].handler = ondemand->valhalla->dispatcher;
-  pause[i++].handler = ondemand->valhalla->dbmanager;
+  pause[i++].handler = VH_HANDLE->parser;
+  pause[i++].handler = VH_HANDLE->dispatcher;
+  pause[i++].handler = VH_HANDLE->dbmanager;
 
   do
   {
@@ -152,8 +164,8 @@ ondemand_thread (void *arg)
 
     if (lstat (file, &st))
     {
-      valhalla_log (VALHALLA_MSG_WARNING,
-                    "[%s] File %s unavailable", __FUNCTION__, file);
+      vh_log (VALHALLA_MSG_WARNING,
+              "[%s] File %s unavailable", __FUNCTION__, file);
       free (file);
       continue;
     }
@@ -162,13 +174,16 @@ ondemand_thread (void *arg)
      * Ignore the ondemand query if the file is already fully available.
      * The ENDED event is sent by this function if the return value is 1.
      */
-    if (vh_dbmanager_file_complete (ondemand->valhalla->dbmanager,
+    if (vh_dbmanager_file_complete (VH_HANDLE->dbmanager,
                                     file, (int64_t) st.st_mtime))
     {
       free (file);
       continue;
     }
 
+    VH_STATS_COUNTER_INC (ondemand->st_cnt);
+    VH_STATS_TIMER_START (ondemand->st_tmr);
+
     /*
      * On-demand engaged, all threads must be in pause mode.
      * After this loop, we are sure that "all" threads are in waiting list.
@@ -205,26 +220,27 @@ ondemand_thread (void *arg)
     }
     /* Check if the file is available and consistent. */
     else if (S_ISREG (st.st_mode)
-             && !vh_scanner_suffix_cmp (ondemand->valhalla->scanner, file))
+             && !vh_scanner_suffix_cmp (VH_HANDLE->scanner, file))
     {
-      int outofpath =
-        !!vh_scanner_path_cmp (ondemand->valhalla->scanner, file);
+      int outofpath = !!vh_scanner_path_cmp (VH_HANDLE->scanner, file);
 
-      fdata = vh_file_data_new (file, st.st_mtime, outofpath, OD_TYPE_NEW,
+      fdata = vh_file_data_new (file, &st, outofpath, OD_TYPE_NEW,
                                 FIFO_QUEUE_PRIORITY_HIGH, STEP_PARSING);
       if (fdata)
-        vh_dbmanager_action_send (ondemand->valhalla->dbmanager,
+        vh_dbmanager_action_send (VH_HANDLE->dbmanager,
                                   fdata->priority, ACTION_DB_NEWFILE, fdata);
     }
     else
-      valhalla_log (VALHALLA_MSG_WARNING,
-                    "[%s] File %s unsupported", __FUNCTION__, file);
+      vh_log (VALHALLA_MSG_WARNING,
+              "[%s] File %s unsupported", __FUNCTION__, file);
 
     free (file);
 
     /* Wake up threads */
     for (i = 0; i < ARRAY_NB_ELEMENTS (pause); i++)
       pause[i].fct_pause (pause[i].handler);
+
+    VH_STATS_TIMER_STOP (ondemand->st_tmr);
   }
   while (!ondemand_is_stopped (ondemand));
 
@@ -237,7 +253,7 @@ vh_ondemand_run (ondemand_t *ondemand, int priority)
   int res = ONDEMAND_SUCCESS;
   pthread_attr_t attr;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!ondemand)
     return ONDEMAND_ERROR_HANDLER;
@@ -262,7 +278,7 @@ vh_ondemand_run (ondemand_t *ondemand, int priority)
 fifo_queue_t *
 vh_ondemand_fifo_get (ondemand_t *ondemand)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!ondemand)
     return NULL;
@@ -273,7 +289,7 @@ vh_ondemand_fifo_get (ondemand_t *ondemand)
 void
 vh_ondemand_stop (ondemand_t *ondemand, int f)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!ondemand)
     return;
@@ -299,7 +315,7 @@ vh_ondemand_stop (ondemand_t *ondemand, int f)
 void
 vh_ondemand_uninit (ondemand_t *ondemand)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!ondemand)
     return;
@@ -310,12 +326,35 @@ vh_ondemand_uninit (ondemand_t *ondemand)
   free (ondemand);
 }
 
+static void
+ondemand_stats_dump (vh_stats_t *stats, void *data)
+{
+  ondemand_t *ondemand = data;
+  uint64_t total;
+  float time;
+
+  if (!stats || !ondemand)
+    return;
+
+  vh_log (VALHALLA_MSG_INFO,
+          "==================================================");
+  vh_log (VALHALLA_MSG_INFO, "Statistics dump (" STATS_GROUP ")");
+  vh_log (VALHALLA_MSG_INFO,
+          "~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
+
+  total = vh_stats_counter_read (ondemand->st_cnt);
+  time  = vh_stats_timer_read (ondemand->st_tmr) / 1000000000.0;
+  vh_log (VALHALLA_MSG_INFO,
+          "Queries    | %6"PRIu64"  %7.2f sec  %7.2f sec/file",
+          total, time, total ? time / total : 0.0);
+}
+
 ondemand_t *
 vh_ondemand_init (valhalla_t *handle)
 {
   ondemand_t *ondemand;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle)
     return NULL;
@@ -328,10 +367,17 @@ vh_ondemand_init (valhalla_t *handle)
   if (!ondemand->fifo)
     goto err;
 
-  ondemand->valhalla = handle;
+  ondemand->valhalla = handle; /* VH_HANDLE */
 
   pthread_mutex_init (&ondemand->mutex_run, NULL);
 
+  /* init statistics */
+  vh_stats_grp_add (handle->stats, STATS_GROUP, ondemand_stats_dump, ondemand);
+  ondemand->st_cnt =
+    vh_stats_grp_counter_add (handle->stats, STATS_GROUP, STATS_GROUP, NULL);
+  ondemand->st_tmr =
+    vh_stats_grp_timer_add (handle->stats, STATS_GROUP, STATS_GROUP, NULL);
+
   return ondemand;
 
  err:
@@ -343,7 +389,7 @@ void
 vh_ondemand_action_send (ondemand_t *ondemand,
                          fifo_queue_prio_t prio, int action, void *data)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!ondemand)
     return;
