@@ -22,6 +22,23 @@
 #ifndef VALHALLA_SQL_STATEMENTS_H
 #define VALHALLA_SQL_STATEMENTS_H
 
+/*
+ * PRIi64 is replaced by I64i with inttypes.h provided by MinGW. These SQL
+ * statements are built with sqlite3_snprintf() [1]. This one does not use
+ * the functions from MSVC, then I64i is not known by SQLite.
+ *
+ * Note that %lli is supported only by sqlite3_snprintf(). With the *printf
+ * functions of MSVC, %ll is considered like %l. It is for this reason that
+ * MinGW uses %I64i instead of %lli in inttypes.h.
+ *
+ * [1]: see SQL_CONCAT() in database.c
+ */
+#ifdef _WIN32
+#define VH_I64 "lli"
+#else
+#define VH_I64 PRIi64
+#endif /* !_WIN32 */
+
 /******************************************************************************/
 /*                                                                            */
 /*                                 Controls                                   */
@@ -72,7 +89,15 @@
 #define CREATE_TABLE_DATA                                 \
  "CREATE TABLE IF NOT EXISTS data ( "                     \
    "data_id          INTEGER PRIMARY KEY AUTOINCREMENT, " \
-   "data_value       TEXT    NOT NULL UNIQUE "            \
+   "data_value       TEXT    NOT NULL UNIQUE, "           \
+   "_lang_id         INTEGER NULL "                       \
+ ");"
+
+#define CREATE_TABLE_LANG                                 \
+ "CREATE TABLE IF NOT EXISTS lang ( "                     \
+   "lang_id          INTEGER PRIMARY KEY AUTOINCREMENT, " \
+   "lang_short       TEXT    NOT NULL UNIQUE, "           \
+   "lang_long        TEXT    NULL "                       \
  ");"
 
 #define CREATE_TABLE_GROUP                                \
@@ -103,6 +128,7 @@
    "data_id          INTEGER NOT NULL, "                  \
    "_grp_id          INTEGER NOT NULL, "                  \
    "external         INTEGER NOT NULL, "                  \
+   "priority__       INTEGER NOT NULL, "                  \
    "PRIMARY KEY (file_id, meta_id, data_id) "             \
  ");"
 
@@ -131,36 +157,36 @@
  "CREATE INDEX IF NOT EXISTS "    \
  "outofpath_idx ON file (outofpath__);"
 
-/* Unique */
-
-#define CREATE_INDEX_FILE_PATH         \
- "CREATE UNIQUE INDEX IF NOT EXISTS "  \
- "file_path_idx ON file (file_path);"
-
-#define CREATE_INDEX_TYPE_NAME         \
- "CREATE UNIQUE INDEX IF NOT EXISTS "  \
- "type_name_idx ON type (type_name);"
-
-#define CREATE_INDEX_META_NAME         \
- "CREATE UNIQUE INDEX IF NOT EXISTS "  \
- "meta_name_idx ON meta (meta_name);"
-
-#define CREATE_INDEX_DATA_VALUE        \
- "CREATE UNIQUE INDEX IF NOT EXISTS "  \
- "data_value_idx ON data (data_value);"
-
-#define CREATE_INDEX_GROUP_NAME        \
- "CREATE UNIQUE INDEX IF NOT EXISTS "  \
- "grp_name_idx ON grp (grp_name);"
-
-#define CREATE_INDEX_GRABBER_NAME      \
- "CREATE UNIQUE INDEX IF NOT EXISTS "  \
- "grabber_name_idx ON grabber (grabber_name);"
-
-#define CREATE_INDEX_ASSOC             \
- "CREATE INDEX IF NOT EXISTS "         \
+#define CREATE_INDEX_ASSOC        \
+ "CREATE INDEX IF NOT EXISTS "    \
  "assoc_idx ON assoc_file_metadata (meta_id, data_id);"
 
+#define CREATE_INDEX_FK_FILE      \
+ "CREATE INDEX IF NOT EXISTS "    \
+ "file_fk_idx ON dlcontext (_file_id);"
+
+#define CREATE_INDEX_FK_ASSOC     \
+ "CREATE INDEX IF NOT EXISTS "    \
+ "grp_fk_idx ON assoc_file_metadata (_grp_id);"
+
+/******************************************************************************/
+/*                                                                            */
+/*                                 Updater                                    */
+/*                                                                            */
+/******************************************************************************/
+
+/* Updater from 1 to 2 */
+
+/* The first language is always the "undef" then the ID is 1. */
+#define DB_UPDATER_FROM_1_TO_2_A            \
+ "ALTER TABLE data "                        \
+ "ADD COLUMN _lang_id INTEGER DEFAULT 1;"
+
+/* A priority of 0 is NORMAL. */
+#define DB_UPDATER_FROM_1_TO_2_B            \
+ "ALTER TABLE assoc_file_metadata "         \
+ "ADD COLUMN priority__ INTEGER DEFAULT 0;"
+
 /******************************************************************************/
 /*                                                                            */
 /*                                  Select                                    */
@@ -172,7 +198,8 @@
 #define SELECT_FILE_FROM                                               \
  "SELECT file.file_id, assoc._grp_id, "                                \
         "meta.meta_id, data.data_id, "                                 \
-        "meta.meta_name, data.data_value, assoc.external "             \
+        "meta.meta_name, data.data_value, "                            \
+        "data._lang_id, assoc.external "                               \
  "FROM (( "                                                            \
      "file INNER JOIN assoc_file_metadata AS assoc "                   \
      "ON file.file_id = assoc.file_id "                                \
@@ -182,10 +209,10 @@
  "ON assoc.meta_id = meta.meta_id "
 
 #define SELECT_FILE_END \
- ";"
+ "ORDER BY assoc.priority__;"
 
 #define SELECT_FILE_WHERE_FILE_ID \
- "file.file_id = %"PRIi64" "
+ "file.file_id = %"VH_I64" "
 #define SELECT_FILE_WHERE_FILE_PATH \
  "file.file_path = '%q' "
 
@@ -199,13 +226,14 @@
  "ORDER BY file_id;"
 
 #define SELECT_LIST_WHERE_TYPE_ID \
- "_type_id = %"PRIi64" "
+ "_type_id = %"VH_I64" "
 
 /* Metadata list selection */
 
 #define SELECT_LIST_METADATA_FROM                         \
  "SELECT meta.meta_id, data.data_id, "                    \
         "meta.meta_name, data.data_value, "               \
+        "data._lang_id, "                                 \
         "assoc._grp_id, assoc.external "                  \
  "FROM ( "                                                \
    "data INNER JOIN assoc_file_metadata AS assoc "        \
@@ -213,6 +241,13 @@
  ") INNER JOIN meta "                                     \
  "ON assoc.meta_id = meta.meta_id "
 
+#define SELECT_LIST_METADATA_WHERE_TYPE_ID  \
+ "assoc.file_id IN ( "                      \
+   "SELECT file_id "                        \
+   "FROM file "                             \
+   "WHERE _type_id = %"VH_I64" "            \
+ ") "
+
 #define SELECT_LIST_METADATA_END          \
  "GROUP BY assoc.meta_id, assoc.data_id " \
  "ORDER BY data.data_value;"
@@ -243,13 +278,17 @@
 #define SELECT_LIST_WHERE_META_NAME \
  "meta.meta_name = '%q' "
 #define SELECT_LIST_WHERE_META_ID \
- "meta.meta_id = %"PRIi64" "
+ "meta.meta_id = %"VH_I64" "
 #define SELECT_LIST_WHERE_DATA_NAME \
  "data.data_value = '%q' "
 #define SELECT_LIST_WHERE_DATA_ID \
- "data.data_id = %"PRIi64" "
+ "data.data_id = %"VH_I64" "
+#define SELECT_LIST_WHERE_LANG_ID \
+ "data._lang_id = %"VH_I64" "
 #define SELECT_LIST_WHERE_GROUP_ID \
- "assoc._grp_id = %"PRIi64" "
+ "assoc._grp_id = %"VH_I64" "
+#define SELECT_LIST_WHERE_PRIORITY \
+ "assoc.priority__ <= %i " /* << highest,  >> lowest */
 
 /* Internal */
 
@@ -283,6 +322,11 @@
  "FROM data "            \
  "WHERE data_value = ?;"
 
+#define SELECT_LANG_ID   \
+ "SELECT lang_id "       \
+ "FROM lang "            \
+ "WHERE lang_short = ?;"
+
 #define SELECT_GROUP_ID  \
  "SELECT grp_id "        \
  "FROM grp "             \
@@ -298,6 +342,15 @@
  "FROM file "            \
  "WHERE file_path = ?;"
 
+#define SELECT_FILE_ID_BY_META                        \
+ "SELECT file.file_id "                               \
+ "FROM ( "                                            \
+   "file INNER JOIN assoc_file_metadata AS assoc "    \
+   "ON file.file_id = assoc.file_id "                 \
+ ") INNER JOIN meta "                                 \
+ "ON assoc.meta_id = meta.meta_id "                   \
+ "WHERE file.file_path = ? AND meta.meta_name = ?;"
+
 #define SELECT_FILE_ID_BY_METADATA                    \
  "SELECT file.file_id "                               \
  "FROM (( "                                           \
@@ -319,6 +372,11 @@
  "FROM file "                     \
  "WHERE checked__ = 0 AND outofpath__ = 0;"
 
+#define SELECT_FILE_OUTOFPATH_SET \
+ "SELECT file_path "              \
+ "FROM file "                     \
+ "WHERE outofpath__ = 1;"
+
 #define SELECT_FILE_GRABBER_NAME                      \
  "SELECT grabber.grabber_name "                       \
  "FROM ( "                                            \
@@ -341,7 +399,7 @@
 /******************************************************************************/
 
 #define INSERT_INFO                   \
- "INSERT "                            \
+ "INSERT OR REPLACE "                 \
  "INTO info (info_name, info_value) " \
  "VALUES (?, ?);"
 
@@ -366,8 +424,15 @@
 
 #define INSERT_DATA        \
  "INSERT "                 \
- "INTO data (data_value) " \
- "VALUES (?);"
+ "INTO data (data_value, " \
+ "           _lang_id) "   \
+ "VALUES (?, ?);"
+
+#define INSERT_LANG        \
+ "INSERT "                 \
+ "INTO lang (lang_short, " \
+ "           lang_long) "  \
+ "VALUES (?, ?);"
 
 #define INSERT_GROUP       \
  "INSERT "                 \
@@ -386,8 +451,9 @@
 
 #define INSERT_ASSOC_FILE_METADATA                                          \
  "INSERT "                                                                  \
- "INTO assoc_file_metadata (file_id, meta_id, data_id, _grp_id, external) " \
- "VALUES (?, ?, ?, ?, ?);"
+ "INTO assoc_file_metadata (file_id, meta_id, data_id, "                    \
+                            "_grp_id, external, priority__) "               \
+ "VALUES (?, ?, ?, ?, ?, ?);"
 
 #define INSERT_ASSOC_FILE_GRABBER                                 \
  "INSERT "                                                        \
@@ -400,11 +466,6 @@
 /*                                                                            */
 /******************************************************************************/
 
-#define UPDATE_INFO     \
- "UPDATE info "         \
- "SET info_value = ? "  \
- "WHERE info_name = ?;"
-
 #define UPDATE_FILE          \
  "UPDATE file "              \
  "SET file_mtime      = ?, " \
@@ -431,9 +492,25 @@
 #define UPDATE_ASSOC_FILE_METADATA \
  "UPDATE assoc_file_metadata "     \
  "SET _grp_id  = ?, "              \
- "    external = ?  "              \
+ "    external = ?, "              \
+ "    priority__ = ? "             \
  "WHERE file_id = ? AND meta_id = ? AND data_id = ?; "
 
+#define UPDATE_ASSOC_FILE_MD_P    \
+ "UPDATE assoc_file_metadata "    \
+ "SET priority__ = ? "            \
+ "WHERE file_id = ?;"
+
+#define UPDATE_ASSOC_FILE_MD_PM   \
+ "UPDATE assoc_file_metadata "    \
+ "SET priority__ = ? "            \
+ "WHERE file_id = ? AND meta_id = ?;"
+
+#define UPDATE_ASSOC_FILE_MD_PMD  \
+ "UPDATE assoc_file_metadata "    \
+ "SET priority__ = ? "            \
+ "WHERE file_id = ? AND meta_id = ? AND data_id = ?;"
+
 /******************************************************************************/
 /*                                                                            */
 /*                                  Delete                                    */
