@@ -22,6 +22,14 @@
 #ifndef VALHALLA_URL_UTILS_H
 #define VALHALLA_URL_UTILS_H
 
+enum url_errno {
+  URL_ERROR_PARAMS    = -4,
+  URL_ERROR_FILE      = -3,
+  URL_ERROR_ABORT     = -2,
+  URL_ERROR_TRANSFER  = -1,
+  URL_SUCCESS         =  0,
+};
+
 typedef struct url_data_s {
   int status;
   char *buffer;
@@ -29,14 +37,21 @@ typedef struct url_data_s {
 } url_data_t;
 
 typedef void url_t;
+typedef struct url_ctl_s url_ctl_t;
 
 void vh_url_global_init (void);
 void vh_url_global_uninit (void);
 
-url_t *vh_url_new (void);
+url_t *vh_url_new (url_ctl_t *abort);
 void vh_url_free (url_t *url);
 url_data_t vh_url_get_data (url_t *handler, char *url);
 char *vh_url_escape_string (url_t *handler, const char *buf);
 int vh_url_save_to_disk (url_t *handler, char *src, char *dst);
 
+url_ctl_t *vh_url_ctl_new (void);
+void vh_url_ctl_free (url_ctl_t *url_ctl);
+void vh_url_ctl_abort (url_ctl_t *url_ctl);
+
+#define MAX_URL_SIZE 1024
+
 #endif /* VALHALLA_URL_UTILS_H */
