@@ -1,6 +1,6 @@
 /*
  * GeeXboX Valhalla: tiny media scanner API.
- * Copyright (C) 2009 Mathieu Schroeter <mathieu.schroeter@gamesover.ch>
+ * Copyright (C) 2009-2010 Mathieu Schroeter <mathieu.schroeter@gamesover.ch>
  *
  * This file is part of libvalhalla.
  *
@@ -37,6 +37,9 @@
 #include "ondemand.h"
 #include "event_handler.h"
 #include "utils.h"
+#include "osdep.h"
+#include "stats.h"
+#include "metadata.h"
 #include "logs.h"
 
 #ifdef USE_GRABBER
@@ -45,6 +48,15 @@
 #include "url_utils.h"
 #endif /* USE_GRABBER */
 
+static int g_preinit;
+static pthread_mutex_t g_preinit_mutex = PTHREAD_MUTEX_INITIALIZER;
+
+
+unsigned int
+libvalhalla_version (void)
+{
+  return LIBVALHALLA_VERSION_INT;
+}
 
 /******************************************************************************/
 /*                                                                            */
@@ -52,6 +64,64 @@
 /*                                                                            */
 /******************************************************************************/
 
+static void
+queue_cleanup (fifo_queue_t *queue)
+{
+  int e;
+  void *data;
+
+  vh_fifo_queue_push (queue,
+                      FIFO_QUEUE_PRIORITY_NORMAL, ACTION_CLEANUP_END, NULL);
+
+  do
+  {
+    e = ACTION_NO_OPERATION;
+    data = NULL;
+    vh_fifo_queue_pop (queue, &e, &data);
+
+    switch (e)
+    {
+    default:
+      break;
+
+    case ACTION_DB_INSERT_P:
+    case ACTION_DB_INSERT_G:
+    case ACTION_DB_UPDATE_P:
+    case ACTION_DB_UPDATE_G:
+    case ACTION_DB_END:
+    case ACTION_DB_NEWFILE:
+      if (data)
+        vh_file_data_free (data);
+      break;
+
+    case ACTION_DB_EXT_INSERT:
+    case ACTION_DB_EXT_UPDATE:
+    case ACTION_DB_EXT_DELETE:
+    case ACTION_DB_EXT_PRIORITY:
+      if (data)
+        vh_dbmanager_extmd_free (data);
+      break;
+
+    case ACTION_OD_ENGAGE:
+    case ACTION_EH_EVENTGL:
+      if (data)
+        free (data);
+      break;
+
+    case ACTION_EH_EVENTOD:
+      if (data)
+        vh_event_handler_od_free (data);
+      break;
+
+    case ACTION_EH_EVENTMD:
+      if (data)
+        vh_event_handler_md_free (data);
+      break;
+    }
+  }
+  while (e != ACTION_CLEANUP_END);
+}
+
 static void
 valhalla_mrproper (valhalla_t *handle)
 {
@@ -70,7 +140,7 @@ valhalla_mrproper (valhalla_t *handle)
     vh_event_handler_fifo_get (handle->event_handler),
   };
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle)
     return;
@@ -136,7 +206,9 @@ valhalla_mrproper (valhalla_t *handle)
       case ACTION_DB_EXT_INSERT:
       case ACTION_DB_EXT_UPDATE:
       case ACTION_DB_EXT_DELETE:
-      case ACTION_EH_EVENT:
+      case ACTION_EH_EVENTOD:
+      case ACTION_EH_EVENTMD:
+      case ACTION_EH_EVENTGL:
         vh_fifo_queue_push (fifo_o, FIFO_QUEUE_PRIORITY_NORMAL, e, data);
         break;
 
@@ -151,14 +223,96 @@ valhalla_mrproper (valhalla_t *handle)
   vh_dbmanager_db_end_transaction (handle->dbmanager);
 #endif /* USE_GRABBER */
 
-  vh_queue_cleanup (fifo_o);
+  queue_cleanup (fifo_o);
   vh_fifo_queue_free (fifo_o);
 
   /* On-demand queue must be handled separately. */
   fifo_o = vh_ondemand_fifo_get (handle->ondemand);
   if (!fifo_o)
     return;
-  vh_queue_cleanup (fifo_o);
+  queue_cleanup (fifo_o);
+}
+
+int
+valhalla_config_set_orig (valhalla_t *handle, valhalla_cfg_t conf, ...)
+{
+  int res = 0;
+  const void *p1 = NULL, *p2 = NULL;
+  int i = 0;
+
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!handle)
+    return -1;
+
+  if (conf >= (1 << VH_CFG_RANGE))
+  {
+    va_list ap;
+
+    va_start (ap, conf);
+
+    if (conf & VH_VOIDP_T)
+      p1 = va_arg (ap, void *);
+    if (conf & VH_INT_T)
+      i  = va_arg (ap, int);
+    if (conf & VH_VOIDP_2_T)
+      p2 = va_arg (ap, void *);
+
+    if (va_arg (ap, int) != ~0) /* check for safeguard */
+    {
+      vh_log (VALHALLA_MSG_CRITICAL,
+              "unrecoverable error with valhalla_config_set(), conf = %#x, "
+              "it is probably a bad use of this function", conf);
+      res = -2;
+    }
+
+    va_end (ap);
+  }
+
+  if (res)
+    return res;
+
+  switch (conf)
+  {
+#ifdef USE_GRABBER
+  case VALHALLA_CFG_DOWNLOADER_DEST:
+    vh_downloader_destination_set (handle->downloader, (valhalla_dl_t) i, p1);
+    break;
+
+  case VALHALLA_CFG_GRABBER_PRIORITY:
+    vh_grabber_priority_set (handle->grabber,
+                             p1, (valhalla_metadata_pl_t) i, p2);
+    break;
+
+  case VALHALLA_CFG_GRABBER_STATE:
+    if (p1)
+      vh_grabber_state_set (handle->grabber, p1, i);
+    break;
+#endif /* USE_GRABBER */
+
+  case VALHALLA_CFG_PARSER_KEYWORD:
+    if (p1)
+      vh_parser_bl_keyword_add (handle->parser, p1);
+    break;
+
+  case VALHALLA_CFG_SCANNER_PATH:
+    if (p1)
+      vh_scanner_path_add (handle->scanner, p1, i);
+    break;
+
+  case VALHALLA_CFG_SCANNER_SUFFIX:
+    if (p1)
+      vh_scanner_suffix_add (handle->scanner, p1);
+    break;
+
+  default:
+    vh_log (VALHALLA_MSG_WARNING,
+            "%s: unsupported option %#x", __FUNCTION__, conf);
+    res = -1;
+    break;
+  }
+
+  return res;
 }
 
 void
@@ -166,9 +320,9 @@ valhalla_wait (valhalla_t *handle)
 {
   const int f = STOP_FLAG_REQUEST | STOP_FLAG_WAIT;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
-  if (!handle)
+  if (!handle || handle->noscan)
     return;
 
   vh_scanner_wait (handle->scanner);
@@ -182,6 +336,8 @@ valhalla_wait (valhalla_t *handle)
   vh_downloader_stop (handle->downloader, f);
 #endif /* USE_GRABBER */
   vh_event_handler_stop (handle->event_handler, f);
+
+  handle->fstop = 1;
 }
 
 static void
@@ -189,21 +345,16 @@ valhalla_force_stop (valhalla_t *handle)
 {
   unsigned int i;
 
-  /*
-   * Because the ondemand thread can send other threads like the dbmanager,
-   * the parser, the grabber, etc, ... in waiting list, this one must be
-   * the first to be stopped.
-   *
-   * TODO: a better way is to improve the pause functionality in order
-   *       to unpause a thread when its stop function is called.
-   */
-  vh_ondemand_stop (handle->ondemand, STOP_FLAG_REQUEST | STOP_FLAG_WAIT);
+  vh_log (VALHALLA_MSG_WARNING,
+          "%s: This can take a time, please be patient", __FUNCTION__);
 
   for (i = 0; i < 2; i++)
   {
     int f = !i ? STOP_FLAG_REQUEST : STOP_FLAG_WAIT;
 
-    vh_scanner_stop (handle->scanner, f);
+    vh_ondemand_stop (handle->ondemand, f);
+    if (!handle->noscan)
+      vh_scanner_stop (handle->scanner, f);
     vh_dbmanager_stop (handle->dbmanager, f);
     vh_dispatcher_stop (handle->dispatcher, f);
     vh_parser_stop (handle->parser, f);
@@ -212,20 +363,39 @@ valhalla_force_stop (valhalla_t *handle)
     vh_downloader_stop (handle->downloader, f);
 #endif /* USE_GRABBER */
     vh_event_handler_stop (handle->event_handler, f);
+
+#ifdef USE_GRABBER
+    /* abort all cURL transfers as fast as possible */
+    if (f == STOP_FLAG_REQUEST)
+      vh_url_ctl_abort (handle->url_ctl);
+#endif /* USE_GRABBER */
   }
 
   valhalla_mrproper (handle);
+
+  handle->fstop = 1;
 }
 
 void
 valhalla_uninit (valhalla_t *handle)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, "%s: begin", __FUNCTION__);
+  int preinit;
+
+  vh_log (VALHALLA_MSG_VERBOSE, "%s: begin", __FUNCTION__);
 
   if (!handle)
     return;
 
-  valhalla_force_stop (handle);
+  pthread_mutex_lock (&g_preinit_mutex);
+  preinit = --g_preinit;
+  pthread_mutex_unlock (&g_preinit_mutex);
+
+  if (!handle->fstop)
+    valhalla_force_stop (handle);
+
+  /* dump all statistics */
+  vh_stats_dump (handle->stats, NULL);
+  vh_stats_debug_dump (handle->stats);
 
   vh_ondemand_uninit (handle->ondemand);
   vh_scanner_uninit (handle->scanner);
@@ -240,23 +410,28 @@ valhalla_uninit (valhalla_t *handle)
 
 #if USE_GRABBER
   vh_url_global_uninit ();
+  vh_url_ctl_free (handle->url_ctl);
 #endif /* USE_GRABBER */
 
 #ifdef USE_LAVC
-  av_lockmgr_register (NULL);
+  if (!preinit)
+    av_lockmgr_register (NULL);
 #endif /* USE_LAVC */
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, "%s: end", __FUNCTION__);
+  vh_stats_free (handle->stats);
+
+  vh_log (VALHALLA_MSG_VERBOSE, "%s: end", __FUNCTION__);
 
   free (handle);
 }
 
 int
-valhalla_run (valhalla_t *handle, int loop, uint16_t timeout, int priority)
+valhalla_run (valhalla_t *handle,
+              int loop, uint16_t timeout, uint16_t delay, int priority)
 {
   int res;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle)
     return VALHALLA_ERROR_HANDLER;
@@ -273,9 +448,17 @@ valhalla_run (valhalla_t *handle, int loop, uint16_t timeout, int priority)
       return VALHALLA_ERROR_THREAD;
   }
 
-  res = vh_scanner_run (handle->scanner, loop, timeout, priority);
+  res = vh_scanner_run (handle->scanner, loop, timeout, delay, priority);
   if (res)
-    return VALHALLA_ERROR_THREAD;
+  {
+    if (res == SCANNER_ERROR_PATH)
+    {
+      handle->noscan = 1;
+      vh_log (VALHALLA_MSG_INFO , "no path defined, scanner disabled");
+    }
+    else
+      return VALHALLA_ERROR_THREAD;
+  }
 
   res = vh_dbmanager_run (handle->dbmanager, priority);
   if (res)
@@ -306,91 +489,71 @@ valhalla_run (valhalla_t *handle, int loop, uint16_t timeout, int priority)
   return VALHALLA_SUCCESS;
 }
 
-void
-valhalla_path_add (valhalla_t *handle, const char *location, int recursive)
+const char *
+valhalla_metadata_group_str (valhalla_meta_grp_t group)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, "%s : %s", __FUNCTION__, location);
-
-  if (!handle || !location)
-    return;
-
-  vh_scanner_path_add (handle->scanner, location, recursive);
-}
-
-void
-valhalla_suffix_add (valhalla_t *handle, const char *suffix)
-{
-  valhalla_log (VALHALLA_MSG_VERBOSE, "%s : %s", __FUNCTION__, suffix);
-
-  if (!handle || !suffix)
-    return;
-
-  vh_scanner_suffix_add (handle->scanner, suffix);
-}
-
-void
-valhalla_bl_keyword_add (valhalla_t *handle, const char *keyword)
-{
-  valhalla_log (VALHALLA_MSG_VERBOSE, "%s : %s", __FUNCTION__, keyword);
-
-  if (!handle || !keyword)
-    return;
-
-  vh_parser_bl_keyword_add (handle->parser, keyword);
-}
-
-void
-valhalla_downloader_dest_set (valhalla_t *handle,
-                              valhalla_dl_t dl, const char *dst)
-{
-  valhalla_log (VALHALLA_MSG_VERBOSE,
-                "%s : (%i) %s", __FUNCTION__, dl, dst ? dst : "?");
-
-  if (!handle || !dst)
-    return;
-
-#ifdef USE_GRABBER
-  vh_downloader_destination_set (handle->downloader, dl, dst);
-#else
-  valhalla_log (VALHALLA_MSG_WARNING,
-                "This function is usable only with grabbing support!");
-#endif /* USE_GRABBER */
-}
-
-void
-valhalla_grabber_state_set (valhalla_t *handle, const char *id, int enable)
-{
-  valhalla_log (VALHALLA_MSG_VERBOSE,
-                "%s : %s (%i)", __FUNCTION__, id ? id : "?", enable);
-
-  if (!handle || !id)
-    return;
-
-#ifdef USE_GRABBER
-  vh_grabber_state_set (handle->grabber, id, enable);
-#else
-  valhalla_log (VALHALLA_MSG_WARNING,
-                "This function is usable only with grabbing support!");
-#endif /* USE_GRABBER */
+  return vh_metadata_group_str (group);
 }
 
 const char *
-valhalla_grabber_list_get (valhalla_t *handle, const char *id)
+valhalla_grabber_next (valhalla_t *handle, const char *id)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, "%s : %s", __FUNCTION__, id ? id : "");
+  vh_log (VALHALLA_MSG_VERBOSE, "%s : %s", __FUNCTION__, id ? id : "");
 
   if (!handle)
     return NULL;
 
 #ifdef USE_GRABBER
-  return vh_grabber_list_get (handle->grabber, id);
+  return vh_grabber_next (handle->grabber, id);
 #else
-  valhalla_log (VALHALLA_MSG_WARNING,
-                "This function is usable only with grabbing support!");
+  vh_log (VALHALLA_MSG_WARNING,
+          "This function is usable only with grabbing support!");
   return NULL;
 #endif /* USE_GRABBER */
 }
 
+valhalla_metadata_pl_t
+valhalla_grabber_priority_read (valhalla_t *handle,
+                                const char *id, const char **meta)
+{
+  vh_log (VALHALLA_MSG_VERBOSE,
+          "%s : %s/%s", __FUNCTION__, id ? id : "", meta && *meta ? *meta : "");
+
+  if (!handle)
+    return 0;
+
+#ifdef USE_GRABBER
+  return vh_grabber_priority_read (handle->grabber, id, meta);
+#else
+  vh_log (VALHALLA_MSG_WARNING,
+          "This function is usable only with grabbing support!");
+  return 0;
+#endif /* USE_GRABBER */
+}
+
+const char *
+valhalla_stats_group_next (valhalla_t *handle, const char *id)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!handle)
+    return NULL;
+
+  return vh_stats_group_next (handle->stats, id);
+}
+
+uint64_t
+valhalla_stats_read_next (valhalla_t *handle, const char *id,
+                          valhalla_stats_type_t type, const char **item)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!handle)
+    return 0;
+
+  return vh_stats_read_next (handle->stats, id, type, item);
+}
+
 void
 valhalla_verbosity (valhalla_verb_t level)
 {
@@ -427,31 +590,58 @@ valhalla_avlock (void **mutex, enum AVLockOp op)
 #endif /* USE_LAVC */
 
 valhalla_t *
-valhalla_init (const char *db,
-               unsigned int parser_nb, int decrapifier, unsigned int commit_int,
-               void (*od_cb) (const char *file, valhalla_event_t e,
-                              const char *id, void *data),
-               void *od_data)
+valhalla_init (const char *db, valhalla_init_param_t *param)
 {
-  static int preinit = 0;
+  int preinit;
   valhalla_t *handle;
+  valhalla_init_param_t p;
+  const valhalla_init_param_t *pp = &p;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  pthread_mutex_lock (&g_preinit_mutex);
+  preinit = g_preinit;
+  pthread_mutex_unlock (&g_preinit_mutex);
+
+  if (!preinit && vh_osdep_init ())
+    return NULL;
 
   if (!db)
     return NULL;
 
+  if (param)
+    pp = param;
+  else
+    memset (&p, 0, sizeof (p));
+
   handle = calloc (1, sizeof (valhalla_t));
   if (!handle)
     return NULL;
 
 #ifdef USE_GRABBER
+  handle->url_ctl = vh_url_ctl_new ();
+  if (!handle->url_ctl)
+    return NULL;
+
   vh_url_global_init ();
 #endif /* USE_GRABBER */
 
-  if (od_cb)
+  handle->stats = vh_stats_new ();
+  if (!handle->stats)
+    goto err;
+
+  if (pp->od_cb || pp->gl_cb || pp->md_cb)
   {
-    handle->event_handler = vh_event_handler_init (handle, od_cb, od_data);
+    event_handler_cb_t cb;
+
+    cb.od_cb = pp->od_cb;
+    cb.gl_cb = pp->gl_cb;
+    cb.md_cb = pp->md_cb;
+    cb.od_data = pp->od_data;
+    cb.gl_data = pp->gl_data;
+    cb.md_data = pp->md_data;
+
+    handle->event_handler = vh_event_handler_init (handle, &cb);
     if (!handle->event_handler)
       goto err;
   }
@@ -460,12 +650,12 @@ valhalla_init (const char *db,
   if (!handle->dispatcher)
     goto err;
 
-  handle->parser = vh_parser_init (handle, parser_nb, decrapifier);
+  handle->parser = vh_parser_init (handle, pp->parser_nb, pp->decrapifier);
   if (!handle->parser)
     goto err;
 
 #ifdef USE_GRABBER
-  handle->grabber = vh_grabber_init (handle);
+  handle->grabber = vh_grabber_init (handle, pp->grabber_nb);
   if (!handle->grabber)
     goto err;
 
@@ -478,7 +668,7 @@ valhalla_init (const char *db,
   if (!handle->scanner)
     goto err;
 
-  handle->dbmanager = vh_dbmanager_init (handle, db, commit_int);
+  handle->dbmanager = vh_dbmanager_init (handle, db, pp->commit_int);
   if (!handle->dbmanager)
     goto err;
 
@@ -494,9 +684,12 @@ valhalla_init (const char *db,
 #endif /* USE_LAVC */
     av_log_set_level (AV_LOG_FATAL);
     av_register_all ();
-    preinit = 1;
   }
 
+  pthread_mutex_lock (&g_preinit_mutex);
+  g_preinit++;
+  pthread_mutex_unlock (&g_preinit_mutex);
+
   return handle;
 
  err:
@@ -507,7 +700,7 @@ valhalla_init (const char *db,
 void
 valhalla_scanner_wakeup (valhalla_t *handle)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle)
     return;
@@ -520,7 +713,7 @@ valhalla_ondemand (valhalla_t *handle, const char *file)
 {
   char *odfile;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle || !file)
     return;
@@ -539,51 +732,76 @@ valhalla_ondemand (valhalla_t *handle, const char *file)
 /*                                                                            */
 /******************************************************************************/
 
-int valhalla_db_metalist_get (valhalla_t *handle,
-                              valhalla_db_item_t *search,
-                              valhalla_db_restrict_t *restriction,
-                              int (*result_cb) (void *data,
-                                                valhalla_db_metares_t *res),
-                              void *data)
+valhalla_db_stmt_t *
+valhalla_db_metalist_get (valhalla_t *handle, valhalla_db_item_t *search,
+                          valhalla_file_type_t filetype,
+                          valhalla_db_restrict_t *restriction)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
-  if (!handle || !search || !result_cb)
-    return -1;
+  if (!handle || !search)
+    return NULL;
 
   return vh_dbmanager_db_metalist_get (handle->dbmanager,
-                                       search, restriction, result_cb, data);
+                                       search, filetype, restriction);
 }
 
-int valhalla_db_filelist_get (valhalla_t *handle,
-                              valhalla_file_type_t filetype,
-                              valhalla_db_restrict_t *restriction,
-                              int (*result_cb) (void *data,
-                                                valhalla_db_fileres_t *res),
-                              void *data)
+const valhalla_db_metares_t *
+valhalla_db_metalist_read (valhalla_t *handle, valhalla_db_stmt_t *vhstmt)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
-  if (!handle || !result_cb)
-    return -1;
+  if (!handle || !vhstmt)
+    return NULL;
 
-  return vh_dbmanager_db_filelist_get (handle->dbmanager,
-                                       filetype, restriction, result_cb, data);
+  return vh_dbmanager_db_metalist_read (handle->dbmanager, vhstmt);
 }
 
-int
-valhalla_db_file_get (valhalla_t *handle,
-                      int64_t id, const char *path,
-                      valhalla_db_restrict_t *restriction,
-                      valhalla_db_filemeta_t **res)
+valhalla_db_stmt_t *
+valhalla_db_filelist_get (valhalla_t *handle, valhalla_file_type_t filetype,
+                          valhalla_db_restrict_t *restriction)
 {
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
-  if (!handle || !res)
-    return -1;
+  if (!handle)
+    return NULL;
 
   return
-    vh_dbmanager_db_file_get (handle->dbmanager, id, path, restriction, res);
+    vh_dbmanager_db_filelist_get (handle->dbmanager, filetype, restriction);
+}
+
+const valhalla_db_fileres_t *
+valhalla_db_filelist_read (valhalla_t *handle, valhalla_db_stmt_t *vhstmt)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!handle || !vhstmt)
+    return NULL;
+
+  return vh_dbmanager_db_filelist_read (handle->dbmanager, vhstmt);
+}
+
+valhalla_db_stmt_t *
+valhalla_db_file_get (valhalla_t *handle, int64_t id, const char *path,
+                      valhalla_db_restrict_t *restriction)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!handle)
+    return NULL;
+
+  return vh_dbmanager_db_file_get (handle->dbmanager, id, path, restriction);
+}
+
+const valhalla_db_metares_t *
+valhalla_db_file_read (valhalla_t *handle, valhalla_db_stmt_t *vhstmt)
+{
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!handle || !vhstmt)
+    return NULL;
+
+  return vh_dbmanager_db_file_read (handle->dbmanager, vhstmt);
 }
 
 /******************************************************************************/
@@ -595,11 +813,11 @@ valhalla_db_file_get (valhalla_t *handle,
 int
 valhalla_db_metadata_insert (valhalla_t *handle, const char *path,
                              const char *meta, const char *data,
-                             valhalla_meta_grp_t group)
+                             valhalla_lang_t lang, valhalla_meta_grp_t group)
 {
   dbmanager_extmd_t *extmd;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle || !path || !meta || !data)
     return -1;
@@ -612,6 +830,7 @@ valhalla_db_metadata_insert (valhalla_t *handle, const char *path,
   extmd->meta  = strdup (meta);
   extmd->data  = strdup (data);
   extmd->group = group;
+  extmd->lang  = lang;
 
   vh_dbmanager_action_send (handle->dbmanager, FIFO_QUEUE_PRIORITY_HIGH,
                             ACTION_DB_EXT_INSERT, extmd);
@@ -621,11 +840,11 @@ valhalla_db_metadata_insert (valhalla_t *handle, const char *path,
 int
 valhalla_db_metadata_update (valhalla_t *handle, const char *path,
                              const char *meta, const char *data,
-                             const char *ndata)
+                             const char *ndata, valhalla_lang_t lang)
 {
   dbmanager_extmd_t *extmd;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle || !path || !meta || !data || !ndata)
     return -1;
@@ -638,6 +857,7 @@ valhalla_db_metadata_update (valhalla_t *handle, const char *path,
   extmd->meta  = strdup (meta);
   extmd->data  = strdup (data);
   extmd->ndata = strdup (ndata);
+  extmd->lang  = lang;
 
   vh_dbmanager_action_send (handle->dbmanager, FIFO_QUEUE_PRIORITY_HIGH,
                             ACTION_DB_EXT_UPDATE, extmd);
@@ -650,7 +870,7 @@ valhalla_db_metadata_delete (valhalla_t *handle, const char *path,
 {
   dbmanager_extmd_t *extmd;
 
-  valhalla_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
 
   if (!handle || !path || !meta || !data)
     return -1;
@@ -667,3 +887,34 @@ valhalla_db_metadata_delete (valhalla_t *handle, const char *path,
                             ACTION_DB_EXT_DELETE, extmd);
   return 0;
 }
+
+int
+valhalla_db_metadata_priority (valhalla_t *handle, const char *path,
+                               const char *meta, const char *data,
+                               valhalla_metadata_pl_t p)
+{
+  dbmanager_extmd_t *extmd;
+
+  vh_log (VALHALLA_MSG_VERBOSE, __FUNCTION__);
+
+  if (!handle || !path)
+    return -1;
+
+  if (!meta && data)
+    return -1;
+
+  extmd = calloc (1, sizeof (dbmanager_extmd_t));
+  if (!extmd)
+    return -1;
+
+  extmd->path = strdup (path);
+  if (meta)
+    extmd->meta = strdup (meta);
+  if (data)
+    extmd->data = strdup (data);
+  extmd->priority = p;
+
+  vh_dbmanager_action_send (handle->dbmanager, FIFO_QUEUE_PRIORITY_HIGH,
+                            ACTION_DB_EXT_PRIORITY, extmd);
+  return 0;
+}
